module _9.result;

import std.conv;
import std.regexp;
import std.string;

class Result{
	static Result parse(char[] id, char[] time = "00:00:01"){
		if(Damage.isDamage(id)){
			return new Damage();
		}
		
		Attack attack = parseAttack(id, time);
		
		if(attack.damage == 0){
			return new Heal();
		}
		
		return attack;
	}
	
	abstract char[] toMessage();
	abstract void apply(inout int enemyHP, inout int cirnoHP);
	
private:
	static Attack parseAttack(char[] id, char[] time){
		char[][] tokens = getTokens(id);
		
		if(BonusAttack.isApplicable(id)){
			int bonus = toInt(time[length - 1..length]);
			return new BonusAttack(tokens, bonus);
		} else {
			return new Attack(tokens);
		}
	}
	
	static char[][] getTokens(char[] id){
		char[][] result;
		uint i = 0;

		char[][] tokens = RegExp(r"(\d+)|(x)", "gi").match(id);
		
		while(!isNumber(tokens[i])){ i++; }
		
		bool multiply = false;
		for(; i < tokens.length; i++){
			if(isNumber(tokens[i])){
				if(multiply){
					result ~= "*";
				}
				result ~= tokens[i];
				multiply = false;
			}else{
				multiply = true;
			}
		}
		
		return result;
	}

	static bool isNumber(char[] token){
		return RegExp(r"\d+").test(token) != 0;
	}
	
}

class Damage : Result{
	static bool isDamage(char[] id){
		return std.regexp.find(id, r"\d") == -1;
	}
	char[] toMessage(){
		return "チルノに⑨のダメージ。";
	}
	override int opEquals(Object o){
		return cast(Damage)o !is null;
	}
	override void apply(inout int enemyHP, inout int cirnoHP){
		cirnoHP -= 9;
	}
}

class Heal : Result{
	char[] toMessage(){
		return "チルノ⑨回復。";
	}
	override int opEquals(Object o){
		return cast(Heal)o !is null;
	}
	override void apply(inout int enemyHP, inout int cirnoHP){
		cirnoHP += 9;
	}
}

class Attack : Result{
	this(char[][] tokens){
		this.tokens = tokens.dup;
	}

	char[] expression(){
		char[] result = tokens[0];

		for(uint i = 1; i < tokens.length; i++){
			char[] token = tokens[i];
			if(isNumber(token)){
				if(toInt(token) != 0){
					result ~= ("+" ~ token);
				}
			} else {
				i++;
				token = tokens[i];
				result ~= ("*" ~ token);
			}
		}
		
		return result;
	}

	int damage(){
		int[] stack;
		stack ~= toInt(tokens[0]);
		for(uint i = 1; i < tokens.length; i++){
			char[] token = tokens[i];
			if(token == "*"){
				stack[length-1] *= toInt(tokens[i+1]);
				i++;
			} else {
				stack ~= toInt(token);
			}
		}
		
		int result = 0;
		foreach(int value; stack){
			result += value;
		}
		
		return result;
	}

	bool isSingleToken(){
		return tokens.length == 1;
	}
	
	bool isNeedExpression(){
		return !isSingleToken;
	}

	char[] toMessage(){
		if(isNeedExpression){
			return format("%s=%dのダメージ。", expression, damage);
		} else {
			return format("%dのダメージ", damage);
		}
	}

	override void apply(inout int enemyHP, inout int cirnoHP){
		enemyHP -= damage;
	}

protected:
	override int opEquals(Object o){
		Attack other = cast(Attack)o;
		return
			(other !is null) &&
			this.tokens == other.tokens;
	}

	char[][] tokens;
}

class BonusAttack : Attack {
	this(char[][] tokens, int bonus){
		super(tokens);
		this.bonus = bonus;
	}

	override char[] toMessage(){
		return "SpellCardBonus!! " ~ super.toMessage();
	}

	override char[] expression(){
		if(isSingleToken()){
			return format("%s*%d", super.expression, bonus);
		}else{
			return format("(%s)*%d", super.expression, bonus);
		}
	}
	
	override bool isNeedExpression(){
		return true;
	}
	
	override int damage(){
		return super.damage * bonus;
	}
	
	static bool isApplicable(char[] id){
		return (std.string.find(id, "9") != -1);
	}
	
private:
	override int opEquals(Object o){
		BonusAttack other = cast(BonusAttack)o;
		return
			super.opEquals(o) &&
			this.bonus == other.bonus;
	}

	int bonus;
}


/** ex: Normal */
unittest{
	assert(Result.parse("IamCirno") == new Damage());
	assert(Result.parse("XEDHYna0") == new Heal);
	assert(Result.parse("0x88jy00") == new Heal());
	assert(Result.parse("r7GF4sa0") == new Attack(["7"[], "4", "0"]));
	assert(Result.parse("ef6oXL5G") == new Attack(["6"[], "*", "5"]));
	assert(Result.parse("8556ERIN") == new Attack(["8556"[]]));
}
/** damage */
unittest{
	assert((new Attack(["7"[], "4", "0"])).damage == 11);
	assert((new Attack(["6"[], "*", "5"])).damage == 30);
	assert((new Attack(["8556"[]])).damage == 8556);
}
/** ex: Spell Card Bonus ! */
unittest{
	assert(Result.parse("3h9jtrw8", "16:16:25") == new BonusAttack(["3"[], "9", "8"], 5));
	assert(Result.parse("Ben49s8k", "19:23:41") == new BonusAttack(["49"[], "8"], 1));
	assert(Result.parse("srEAwe90", "20:41:20") == new Heal());
	assert(Result.parse("77x88x99", "23:59:59") == new BonusAttack(["77"[], "*", "88", "*", "99"], 9));
}
/** damage */
unittest{
	assert((new BonusAttack(["3"[], "9", "8"], 5)).damage == 100);
	assert((new BonusAttack(["49"[], "8"], 1)).damage == 57);
	assert((new BonusAttack(["77"[], "*", "88", "*", "99"], 9)).damage == 6037416);
}
/** other */
unittest{
	assert(Result.parse("Xonlyyyy") == new Damage());
	assert(Result.parse("endWithx") == new Damage());
	assert(Result.parse("100x0aaa") == new Heal());
	assert(Result.parse("3and5xX8") == new Attack(["3"[], "5", "*", "8"]));
	assert(Result.parse("3and5xX9", "00:00:02") == new BonusAttack(["3"[], "5", "*", "9"], 2));
	assert(Result.parse("hoge91aa", "00:00:02") == new BonusAttack(["91"[]], 2));

	assert(Result.parse("6BxQK1DY", "19:06:45") == new Attack(["6"[], "*", "1"]));
}
/** Calc Expression */
unittest{
	assert((new Attack(["41"[], "7", "2"])).expression == "41+7+2");
	assert((new Attack(["7"[], "4", "0"])).expression == "7+4");
	assert((new Attack(["6"[], "*", "5"])).expression == "6*5");
	assert((new Attack(["8556"[]])).expression == "8556");
	assert((new BonusAttack(["3"[], "9", "8"], 5)).expression == "(3+9+8)*5");
	assert((new Attack(["3"[], "5", "*", "8"])).expression == "3+5*8");
	assert((new BonusAttack(["3"[], "5", "*", "9"], 2)).expression == "(3+5*9)*2");
	assert((new Attack(["3"[], "*", "0", "5"])).expression == "3*0+5");
	assert((new Attack(["3"[], "2", "*", "3", "*", "5", "10"])).expression == "3+2*3*5+10");
	assert((new BonusAttack(["9"[]], 2)).expression == "9*2");
}

/** apply */
unittest{
	int enemy = 10000;
	int cirno = 99;

	(new Attack(["10"[], "20"])).apply(enemy, cirno);
	assert(enemy == 9970 && cirno == 99);

	(new Damage()).apply(enemy, cirno);
	assert(enemy == 9970 && cirno == 90);
	
	(new BonusAttack(["10"[], "*", "9"], 3)).apply(enemy, cirno);
	assert(enemy == 9700 && cirno == 90);
	
	(new Heal()).apply(enemy, cirno);
	assert(enemy == 9700 && cirno == 99);
}
